from dataclasses import dataclass
from typing import Optional
import boto3
import re
from urllib.parse import urlparse


@dataclass(frozen=True)
class ConnectionInfo:
    base: str
    path: str


@dataclass(frozen=True)
class LoopConnectionInfo:
    rest_info: ConnectionInfo
    ws_info: ConnectionInfo

    def stream_connection_info(self, stream_id, group=None):
        if group:
            send_path = f"{self.rest_info.path}/stream/{stream_id}/group/{group}"
            connect_path = f"{self.ws_info.path}/stream/{stream_id}/group/{group}"
        else:
            send_path = f"{self.rest_info.path}/stream/{stream_id}"
            connect_path = f"{self.ws_info.path}/stream/{stream_id}"

        return StreamConnectionInfo(
            send_info=ConnectionInfo(self.rest_info.base, send_path),
            connect_info=ConnectionInfo(self.ws_info.base, connect_path),
        )


@dataclass(frozen=True)
class StreamConnectionInfo:
    send_info: Optional[ConnectionInfo]
    connect_info: Optional[ConnectionInfo]


class LoopOperations:
    def __init__(self, connection_info: LoopConnectionInfo, apikey):
        self.connect_info = connection_info
        self.apikey = apikey

    def get_stream(self, stream_id, group=None):
        stream_connection_info = self.connect_info.stream_connection_info(
            stream_id, group
        )
        return StreamOperations(stream_connection_info, self.apikey)


class StreamOperations:
    def __init__(self, connection_info: StreamConnectionInfo, apikey):
        self.connection_info = connection_info
        self.apikey = apikey
        self.socket = None

    async def send(self, str):
        return await self.socket.send(str)

    async def recv(self):
        return await self.socket.recv()

    @property
    def connect_url(self):
        return (
            self.connection_info.connect_info.base
            + self.connection_info.connect_info.path
        )

    @property
    def send_url(self):
        return self.connection_info.send_info.base + self.connection_info.send_info.path



from contextlib import asynccontextmanager
from dataclasses import dataclass
import traceback
import aiohttp
import json
from urllib.parse import urlparse
from .loops import ConnectionInfo, StreamOperations


@dataclass(frozen=True)
class ClusterConnectionInfo:
    configure_info: ConnectionInfo
    connect_info: ConnectionInfo


def create_patch(old_spec, new_spec):
    if new_spec == None:
        return None

    if old_spec == None:
        return new_spec

    patch = {}
    for k, v in new_spec.items():
        if k not in old_spec:
            patch[k] = v
            continue

        if type(v) == list:
            if not all([x == y for x, y in zip(v, old_spec[k])]):
                patch[k] = v
            continue

        if type(v) == dict:
            inner_patch = create_patch(old_spec[k], v)
            if len(inner_patch) > 0:
                patch[k] = inner_patch
            continue

        if v != old_spec[k]:
            patch[k] = v
            continue

    return patch


def merge(old_spec, patch):
    if old_spec == None:
        return patch

    new_spec = old_spec.copy()
    for k, v in patch.items():
        if type(v) == dict:
            new_spec[k] = merge(old_spec[k], v)
        else:
            new_spec[k] = v
    return new_spec


def _remove_scheme(url):
    parsed = urlparse(url)
    if parsed.scheme:
        # Remove the scheme and leading // from the URL
        return parsed._replace(scheme="").geturl()[2:]
    else:
        return url


class ClusterOperations:
    def __init__(self, connection_info: ClusterConnectionInfo, apikey):
        self.connection_info = connection_info
        self.endpoint_url = (
            connection_info.configure_info.base + connection_info.configure_info.path
        )
        self.apikey = apikey
        if apikey:
            self.apikey_param = {"apikey": apikey}
        else:
            self.apikey_param = {}

    async def create_resource(self, config):
        name = config["metadata"]["name"]
        group, version = config["apiVersion"].split("/")
        kind = config["kind"]
        url = f"{self.endpoint_url}/config/{group}/{version}/{kind}"
        async with aiohttp.ClientSession() as session:
            # pass apikey as query parameter
            async with session.post(
                url, json=config, params=self.apikey_param
            ) as response:
                return await response.json()

    async def read_all_resources(self, group, version, kind, name, utctime):
        url = f"{self.endpoint_url}/config"
        params = self.apikey_param.copy()
        if group:
            params["group"] = group
        if version:
            params["version"] = version
        if kind:
            params["kind"] = kind
        if name:
            params["name"] = name
        if utctime:
            params["utctime"] = utctime

        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(url, params=params) as response:
                    return await response.json()
        except Exception as e:
            # print stack track
            traceback.print_exc()
            raise e

    async def read_resource(self, group, version, kind, name):
        url = f"{self.endpoint_url}/config/{group}/{version}/{kind}/{name}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=self.apikey_param) as response:
                return await response.json()

    async def patch_resource(self, config):
        name = config["metadata"]["name"]
        group, version = config["apiVersion"].split("/")
        kind = config["kind"]
        url = f"{self.endpoint_url}/config/{group}/{version}/{kind}/{name}"
        async with aiohttp.ClientSession() as session:
            async with session.patch(
                url, json=config, params=self.apikey_param
            ) as response:
                return await response.read()

    async def update_resource(self, config):
        name = config["metadata"]["name"]
        group, version = config["apiVersion"].split("/")
        kind = config["kind"]
        url = f"{self.endpoint_url}/config/{group}/{version}/{kind}/{name}?create=false"
        async with aiohttp.ClientSession() as session:
            async with session.put(
                url, json=config, params=self.apikey_param
            ) as response:
                return await response.json()

    async def apply_resource(self, config):
        name = config["metadata"]["name"]
        group, version = config["apiVersion"].split("/")
        kind = config["kind"]
        url = f"{self.endpoint_url}/config/{group}/{version}/{kind}/{name}?create=true"
        data = json.dumps(config)
        headers = {"Content-Type": "application/json"}
        async with aiohttp.ClientSession() as session:
            async with session.put(
                url, headers=headers, data=data, params=self.apikey_param
            ) as response:
                text = await response.text()
                # return await response.json()
                return json.loads(text)

    async def delete_resource(self, group, version, kind, name):
        url = f"{self.endpoint_url}/config/{group}/{version}/{kind}/{name}"
        async with aiohttp.ClientSession() as session:
            async with session.delete(url, params=self.apikey_param) as response:
                return await response.json()

    async def read_queue(
        self, group=None, version=None, kind=None, name=None, utctime=None
    ):
        url = f"{self.endpoint_url}/queue"
        params = self.apikey_param.copy()
        if group:
            params["group"] = group
        if version:
            params["version"] = version
        if kind:
            params["kind"] = kind
        if name:
            params["name"] = name
        if utctime:
            params["timestamp"] = utctime

        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as response:
                return await response.json()

    async def lock_resource(self, group, version, kind, name):
        url = f"{self.endpoint_url}/claim/{group}/{version}/{kind}/{name}"

        async with aiohttp.ClientSession() as session:
            async with session.post(url, params=self.apikey_param) as response:
                return await response.json()

    async def unlock_resource(self, group, version, kind, name):
        url = f"{self.endpoint_url}/release-claim/{group}/{version}/{kind}/{name}"
        async with aiohttp.ClientSession() as session:
            async with session.post(url, params=self.apikey_param) as response:
                return await response.json()

    async def resolve_resource(
        self, group, version, kind, name, config, operations, delete=False, force=False
    ):
        url = f"{self.endpoint_url}/resolve-claim/{group}/{version}/{kind}/{name}?force={force}"
        data = {"operations": operations, "delete": delete}
        if config != None:
            data["config"] = config

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, json=data, params=self.apikey_param
            ) as response:
                return await response.json()

    def control_resource(self, group, version, kind, name, validate=False):
        return BaseController(self, group, version, kind, name, validate=validate)


class BaseController:
    def __init__(
        self, config_ops: ClusterOperations, group, version, kind, name, validate=False
    ):
        self.config_ops = config_ops
        self.group = group
        self.version = version
        self.kind = kind
        self.name = name
        self.validate = validate

        self.pending_ops = None
        self.processed_op_ids = set()
        self.observed_op_ids = set()

        self.have_current_config = False
        self.current_config = None
        self.delete_current = False
        self.locked_config_name = None

    async def __aenter__(self):
        await self.acquire_object()

    async def __aexit__(self, exc_type, exc, tb):
        await self.release_current_object()

    async def acquire_object(self):
        initial_ops = await self.config_ops.lock_resource(
            self.group, self.version, self.kind, name=self.name
        )
        self.locked = initial_ops != []

        if initial_ops:
            self.locked_config_name = initial_ops[0]["name"]
        else:
            self.locked_config_name = None

        self.current_config = None
        self.delete_current = False
        self.have_current_config = False
        self.pending_ops = sorted(initial_ops, key=lambda x: x["timestamp"])

    async def release_current_object(self):
        if self.locked_config_name == None:
            return

        await self.config_ops.resolve_resource(
            self.group,
            self.version,
            self.kind,
            self.locked_config_name,
            self.current_config,
            list(self.processed_op_ids),
            delete=self.delete_current,
            force=True,
        )

        self.locked_config_name = None
        self.current_config = None
        self.delete_current = False
        self.have_current_config = False
        self.processed_op_ids = set()

    async def next_operation_batch(self):
        if self.locked_config_name == None:
            return

        force = self.observed_op_ids != self.processed_op_ids
        if force:
            print(
                "Warning: release config lock because not all operations were processed"
            )

        new_ops = await self.config_ops.resolve_resource(
            self.group,
            self.version,
            self.kind,
            self.locked_config_name,
            self.current_config,
            list(self.processed_op_ids),
            delete=self.delete_current,
            force=force,
        )

        self.processed_op_ids = set()

        # If there are new operations for this object, queue them
        if new_ops:
            self.pending_ops = sorted(new_ops, key=lambda x: x["timestamp"])
        elif not self.name:
            # If there are no new operations for this object, check other objects if needed
            self.locked_config_name = None
            await self.acquire_object()
        else:
            self.locked_config_name = None

    def __aiter__(self):
        return self

    async def get_next_operation(self):
        if self.locked_config_name == None:
            raise StopAsyncIteration()

        if not self.pending_ops:
            await self.next_operation_batch()

        if not self.pending_ops:
            raise StopAsyncIteration()

        # Return the next operation
        result = self.pending_ops.pop(0)
        self.observed_op_ids.add(result["id"])
        return PendingOperation(self, result)

    async def __anext__(self):
        result = await self.get_next_operation()
        if not self.validate:
            return result

        while True:
            if await result.validate():
                return result
            await result.reject()
            result = await self.get_next_operation()

    async def get_current_config(self):
        if self.have_current_config == False:
            self.current_config = await self.config_ops.read_resource(
                self.group, self.version, self.kind, self.locked_config_name
            )
            self.have_current_config = True

        return self.current_config

    async def accept(self, pendingOp, new_config=None, delete=None):
        if pendingOp.data["id"] in self.processed_op_ids:
            raise ValueError(f"Operation {pendingOp['id']} already processed")

        self.processed_op_ids.add(pendingOp.data["id"])

        self.current_config = new_config or await pendingOp.new_config()

        if delete == None:
            self.delete_current = pendingOp.data["operation"] == "DELETE"
        else:
            self.delete_current = delete

    async def reject(self, pendingOp):
        if pendingOp.data["id"] in self.processed_op_ids:
            raise ValueError(f"Operation {pendingOp.data['id']} already processed")

        self.processed_op_ids.add(pendingOp.data["id"])


class PendingOperation:
    def __init__(self, controller: BaseController, data):
        self.controller = controller
        self.data = data

    async def old_config(self):
        return await self.controller.get_current_config()

    async def new_config(self):
        if self.data["operation"] == "CREATE":
            return self.data["config"]
        elif self.data["operation"] == "DELETE":
            return None
        elif self.data["operation"] == "PATCH":
            return merge(await self.old_config(), self.data["config"])
        elif self.data["operation"] == "REPLACE":
            return self.data["config"]

    async def patch_config(self):
        return create_patch(await self.old_config(), await self.new_config())

    async def validate(self):
        if self.data["operation"] == "CREATE":
            return await self.old_config() == None
        elif self.data["operation"] == "DELETE":
            return await self.old_config() != None
        elif self.data["operation"] == "PATCH":
            return await self.old_config() != None
        elif self.data["operation"] == "REPLACE":
            if self.data["create"]:
                return True
            else:
                return await self.old_config() != None
        else:
            return False

    async def accept(self, new_config=None, delete=None):
        await self.controller.accept(self, new_config=new_config, delete=delete)

    async def reject(self):
        await self.controller.reject(self)



import boto3
import re
from urllib.parse import urlparse


def split_s3_url(url):
    # Parse the URL
    parsed_url = urlparse(url)

    # Get the netloc (hostname and port)
    netloc = parsed_url.netloc

    # Remove the port number from the netloc
    hostname = netloc.split(":")[0]
    port = netloc.split(":")[1]

    # Construct the endpoint URL
    endpoint_url = f"{parsed_url.scheme}://{hostname}:{port}"

    # Split the path to retrieve the bucket name and object key
    path_parts = parsed_url.path.lstrip("/").split("/", 1)
    bucket_name = path_parts[0]
    object_key = path_parts[1] if len(path_parts) > 1 else None

    return endpoint_url, bucket_name, object_key


class BucketOperations:
    def __init__(self, secret):
        bucket_name = secret["bucketName"]
        endpoint_url = secret["secretS3"]["endpoint"]
        access_key_id = secret["secretS3"]["accessKeyID"]
        access_secret_key = secret["secretS3"]["accessSecretKey"]

        self.bucket_name = bucket_name
        self.client = boto3.client(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=access_secret_key,
            use_ssl=False,
        )

    def put_object(self, key, file_descriptor, metadata={}):
        return self.client.put_object(
            Bucket=self.bucket_name, Key=key, Body=file_descriptor, Metadata=metadata
        )

    def get_object(self, key):
        response = self.client.get_object(Bucket=self.bucket_name, Key=key)
        contents = response["Body"]
        metadata = response["Metadata"]
        return contents, metadata

    def delete_object(self, key):
        return self.client.delete_object(Bucket=self.bucket_name, Key=key)


class PileOperations:
    def __init__(self, bucket, prefix=None):
        self.bucket = bucket
        self.prefix = prefix

    def put(self, key, file_descriptor, metadata={}):
        if self.prefix != None:
            if not key.startswith(self.prefix):
                raise ValueError(f"Key {key} needs to start with prefix {self.prefix}")

        return self.bucket.put_object(key, file_descriptor, metadata=metadata)

    def get(self, key):
        if self.prefix != None:
            if not key.startswith(self.prefix):
                raise ValueError(f"Key {key} needs to start with prefix {self.prefix}")

        contents, metadata = self.bucket.get_object(key)
        return contents, metadata

    def delete(self, key):
        if self.prefix != None:
            if not key.startswith(self.prefix):
                raise ValueError(f"Key {key} needs to start with prefix {self.prefix}")

        return self.bucket.delete_object(Key=key)



import asyncio
from typing import Any
from pydantic import BaseModel


from .itl import Itl
from .clusters import BaseController


class ClusterController:
    def __init__(self, itl: Itl, cluster: str, group: str, version: str, kind: str):
        self.itl = itl
        self.cluster = cluster
        self.group = group
        self.version = version
        self.kind = kind
        self._parents: list[SyncedResources] = []
        self.resources: dict[str, Any] = {}

        print(
            "Loaded resource controller for",
            cluster,
            self.group,
            self.version,
            kind,
        )
        itl.controller(cluster, self.group, self.version, kind)(self.controller)

    async def controller(self, pending: BaseController):
        async for op in pending:
            config = await op.new_config()
            if config == None:
                # Delete the resource
                resource = self._get_resource(pending.name)
                await self.delete_resource(resource)
                self._remove_resource(pending.name)
                await op.accept()
                print("Deleted", f"{self.kind}/{pending.name}")
                continue

            name = config["metadata"]["name"]
            old_resource = self._get_resource(name)

            try:
                if old_resource == None:
                    result = await self.create_resource(config)
                    if result == None:
                        await op.reject()
                        print("Rejected", f"{self.kind}/{pending.name}")
                        continue

                    self._add_resource(pending.name, result)
                    print("Created", f"{self.kind}/{pending.name}")
                else:
                    result = await self.update_resource(old_resource, config)
                    self._add_resource(name, result)
                    print("Reconfigured", f"{self.kind}/{pending.name}")

                await op.accept()

            except Exception as e:
                await op.reject()
                print(f"Failed to load resource {self.kind}/{pending.name}: {e}")

    async def create_resource(self, config):
        raise ValueError("create_resource not implemented")

    async def update_resource(self, resource, config):
        result = await self.create_resource(config)
        if result == None:
            raise ValueError("create_resource returned None for", config)
        return result

    async def delete_resource(self, resource):
        pass

    async def load_existing(self):
        resources = await self.itl.cluster_read_all(
            self.cluster, self.group, self.version, self.kind
        )
        if resources:
            for config in resources:
                try:
                    result = await self.create_resource(config["config"])
                    self._add_resource(config["name"], result)
                    print("Loaded", self.kind, config["name"])
                except Exception as e:
                    print(f'Failed to load resource {config["name"]}: {e}')

    def _register_parent(self, resource_set):
        self._parents.append(resource_set)

    def _add_resource(self, name, resource):
        self.resources[name] = resource
        key = self.group + "/" + self.version + "/" + self.kind + "/" + name
        for parent in self._parents:
            parent._collection[key] = resource

    def _get_resource(self, name):
        return self.resources.get(name)

    def _remove_resource(self, name):
        if name in self.resources:
            del self.resources[name]

        key = self.group + "/" + self.version + "/" + self.kind + "/" + name
        for parent in self._parents:
            if key in parent._collection:
                del parent._collection[key]


class PydanticResourceController(ClusterController):
    def __init__(
        self, resource_cls, itl: Itl, cluster: str, group: str, version: str, kind: str
    ):
        super().__init__(itl, cluster, group, version, kind)
        self.resource_cls = resource_cls

    async def create_resource(self, config):
        if "spec" not in config:
            raise ValueError("Config is missing required key: spec")
        return self.resource_cls(**config["spec"])


class SyncedResources:
    def __init__(self):
        self._collection = {}

    def register(self, itl: Itl, cluster, group, version, kind):
        def decorator(controller_cls):
            if issubclass(controller_cls, ClusterController):
                controller = controller_cls(itl, cluster, group, version, kind)
            elif issubclass(controller_cls, BaseModel):
                controller = PydanticResourceController(
                    controller_cls, itl, cluster, group, version, kind
                )

            self._register_controller(controller)
            itl.onconnect(controller.load_existing)
            return controller_cls

        return decorator

    def _register_controller(self, controller: ClusterController):
        controller._register_parent(self)
        for name, controller in controller.resources.items():
            key = (
                controller.group
                + "/"
                + controller.version
                + "/"
                + controller.kind
                + "/"
                + name
            )
            self._collection[key] = controller

    def __getitem__(self, name):
        return self._collection[name]

    def keys(self):
        return self._collection.keys()

    def values(self):
        return self._collection.values()

    def items(self):
        return self._collection.items()

    def __contains__(self, name):
        return name in self._collection

    def __iter__(self):
        return iter(self._collection)

    def __len__(self):
        return len(self._collection)

    def get(self, name, default=None):
        return self._collection.get(name, default)

    def __setitem__(self, name, value):
        self._collection[name] = value



import argparse
import asyncio
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
import os
from typing import Any, Dict, Generator, List, Optional, Set, Tuple, Union
from urllib.parse import urlparse
import itertools
import re

import fire
import yaml
import httpx
from itllib.clusters import ClusterConnectionInfo

from itllib.loops import LoopConnectionInfo, ConnectionInfo, StreamConnectionInfo

DEFAULT_REMOTES = {
    "https://resources.thatone.ai": {
        "clusters": "https://clusters.thatone.ai",
        "loops": "https://streams.thatone.ai",
    },
    "https://test-resources.thatone.ai": {
        "clusters": "https://test-clusters.thatone.ai",
        "loops": "https://test-streams.thatone.ai",
    },
}


@dataclass(frozen=True)
class ResourceReference:
    kind: str
    name: str

    def __str__(self):
        return f"{self.kind}/{self.name}"


@dataclass(frozen=True)
class LocatedResource:
    path: str
    index: int
    resource: "Resource"


@dataclass(frozen=True)
class LocatedConfig:
    path: str
    index: int
    config: dict


@dataclass(frozen=True)
class ResourceId:
    kind: str
    id: str

    def __str__(self):
        return f"{self.kind}/{self.id}"


@dataclass(frozen=True)
class ResourceIdRef:
    kind: str
    id: Union[str, "DeferredValue"]

    def resolve(self, resolver: "ResourceResolver") -> ResourceId:
        if isinstance(self.id, DeferredValue):
            return ResourceId(self.kind, self.id.link(resolver))
        return ResourceId(self.kind, self.id)


@dataclass(frozen=True)
class ConfigKey:
    reference: ResourceReference
    key: str


CONFIG_KINDS = {
    "ResourceSpec",
    "Remote",
    "Client",
    "ApiKey",
    "Group",
    "Loop",
    "Stream",
    "Cluster",
}


def validate_keys(location: str, config: dict, required_keys: set, optional_keys=set()):
    """Validate that the given config has the required keys and no other keys."""
    keys = set(config.keys())
    if not keys >= required_keys:
        print(f"Error: Missing required keys in {location}: {required_keys - keys}")
        return False
    if not keys <= required_keys | optional_keys:
        print(
            f"Error: Unknown keys in {location}: {keys - required_keys - optional_keys}"
        )
        return False
    return True


CLIENT = httpx.Client()


def get_client() -> httpx.Client:
    global CLIENT
    return CLIENT


def is_itl_resource(config):
    if "apiVersion" not in config or "kind" not in config:
        return False
    if config["apiVersion"] != "thatone.ai/v1":
        return False
    if config["kind"] not in CONFIG_KINDS:
        return False

    return True


def expand_keys(config):
    if config["kind"] != "ResourceSpec":
        yield ResourceReference(config["kind"], config["metadata"]["name"])
        return

    for config_type, config_value in config["spec"].items():
        if config_type == "clients":
            for client in config_value:
                yield ResourceReference("Client", client["name"])
        elif config_type == "groups":
            for group in config_value:
                yield ResourceReference("Group", group["name"])
        elif config_type == "loops":
            for loop in config_value:
                yield ResourceReference("Loop", loop["name"])
        elif config_type == "streams":
            for stream in config_value:
                yield ResourceReference("Stream", stream["name"])
        elif config_type == "clusters":
            for cluster in config_value:
                yield ResourceReference("Cluster", cluster["name"])
        else:
            raise ValueError(f"Unknown config type {config_type}")


class DeferredValue:
    def __str__(self):
        raise NotImplementedError()

    def link(self, resolver: "ResourceResolver"):
        raise NotImplementedError()


class DeferredId(DeferredValue):
    def __init__(self, kind, name, field):
        reference = ResourceReference(kind, name)
        self.configKey = ConfigKey(reference, field)
        self._generated = None

    def link(self, resolver: "ResourceResolver"):
        if self._generated == None:
            self._generated = resolver.get_generated_id(self.configKey)
            id = ResourceId(self.configKey.reference.kind, self._generated)
            resolver.id_to_reference[id] = self.configKey.reference

        return self._generated


class ClientId(DeferredId):
    def __init__(self, name):
        super().__init__("Client", name, "clientId")


class GroupId(DeferredId):
    def __init__(self, name):
        super().__init__("Group", name, "groupId")


class LoopId(DeferredId):
    def __init__(self, name):
        super().__init__("Loop", name, "loopId")


class ClusterId(DeferredId):
    def __init__(self, name):
        super().__init__("Cluster", name, "clusterId")


class Remote(DeferredValue):
    def __init__(self, resourceKey: ResourceReference, remote=None):
        self.resource_reference = resourceKey
        self.remote = remote

    def __str__(self):
        return self.remote or str(self.resource_reference)

    def link(self, resolver: "ResourceResolver"):
        if self.remote:
            return self.remote

        resource = resolver.require_resource_by_reference(self.resource_reference)
        while True:
            # Return the remote if it's known
            remote = resource.remote
            if remote:
                if isinstance(remote, str):
                    return remote
                if isinstance(remote, Remote) and remote.remote:
                    return remote.remote

            # Otherwise, go up a level
            if resource.parent(resolver) == None:
                print("Error: Could not find parent for", resource.reference)
                raise ValueError()
            resource = resource.parent(resolver)


class Resource:
    config: dict[str, Any]
    parent_ref_or_id: Optional[tuple] = None
    kind: str
    name: str
    remote: Union[str, Remote]

    @property
    def api_version(self):
        return self.config["apiVersion"]

    def resource_ids(self) -> Generator[ResourceIdRef, None, None]:
        raise NotImplementedError()

    @property
    def reference(self):
        return ResourceReference(self.kind, self.name)

    def parent(self, resolver: "ResourceResolver"):
        if self.parent_ref_or_id == None:
            return None

        if isinstance(self.parent_ref_or_id, ResourceReference):
            return resolver.get_resource_by_reference(self.parent_ref_or_id)

        if isinstance(self.parent_ref_or_id, ResourceId):
            reference = resolver.id_to_reference[self.parent_ref_or_id]
            return resolver.get_resource_by_reference(reference)

        if isinstance(self.parent_ref_or_id, ResourceIdRef):
            parent_reference = self.parent_ref_or_id.resolve(resolver)
            reference = resolver.id_to_reference[parent_reference]
            return resolver.get_resource_by_reference(reference)

        raise ValueError(f"Invalid parent reference {self.parent_ref_or_id}")

    def get_remote(self, resolver: "ResourceResolver"):
        return resolver.link(self.remote)

    def put_with_key(self, url, json, key=None):
        client = get_client()
        if key:
            params = {"apikey": key}
        else:
            params = {}
        response = client.put(url, json=json, params=params)
        return response

    def apply(self, resolver: "ResourceResolver"):
        # Find all applicable API keys
        api_keys = resolver.apikeys_for_resource(self)

        url = resolver.link(self.remote)
        compiled_config = self.link(resolver)

        response = None
        for apikey in api_keys:
            response = self.put_with_key(
                f"{url}/{self.api_version}/{self.kind}",
                compiled_config["spec"],
                apikey.key(resolver),
            )

            if response.status_code != 401:
                break

            print(
                "Error: API key",
                apikey.key(resolver),
                "is not authorized to apply",
                self.name,
            )
            apikey.valid = False
        else:
            response = self.put_with_key(
                f"{url}/{self.api_version}/{self.kind}", compiled_config["spec"]
            )

        if response == None:
            print(f"Error: no API keys are authorized to apply {self.reference}")
            return False

        elif response.status_code != 200:
            print(
                "- Error: failed to apply",
                self.reference,
                response.status_code,
                response.text,
            )
            return False

        print("applied", self.reference)

        data = response.json()
        if data.get("apikey") != None:
            resolver.store_generated_apikey(self, data["apikey"])

        return True

    def link(self, resolver: "ResourceResolver"):
        compiled_config = self._link(self.config, resolver)
        result = {
            "apiVersion": compiled_config["apiVersion"],
            "kind": compiled_config["kind"],
            "metadata": compiled_config["metadata"],
            "spec": {x: y for x, y in compiled_config["spec"].items() if y},
        }
        return result

    def _link(self, value, resolver: "ResourceResolver"):
        if isinstance(value, DeferredValue):
            return value.link(resolver)
        elif isinstance(value, list):
            return [self._link(x, resolver) for x in value]
        elif isinstance(value, dict):
            return {k: self._link(v, resolver) for k, v in value.items()}
        elif isinstance(value, Resource):
            return value.link(resolver)
        else:
            return value


class ApiKeyResource(Resource):
    valid: bool = True

    def key(self, resolver: "ResourceResolver"):
        return resolver.link(self.config["spec"]["apiKey"])

    def namespace(self, resolver: "ResourceResolver"):
        return resolver.link(self.config["metadata"]["namespace"])

    def __init__(self, config):
        self.kind = "ApiKey"

        if "namespace" not in config["metadata"]:
            print("Error: ApiKey is missing namespace in metadata")
            raise ValueError()
        if not validate_keys("apikey", config.get("spec", {}), {"clientId", "apiKey"}):
            raise ValueError()
        self.config = config
        self.name = config["metadata"]["name"]
        self.remote = config["metadata"]["namespace"]

    def apply(self, *args, **kwargs):
        return True

    def resource_ids(self):
        return
        yield


class ClientResource(Resource):
    def __init__(self, config):
        self.kind = "Client"

        if "namespace" not in config["metadata"]:
            print("Error: Client is missing namespace in metadata")
            raise ValueError()

        if not validate_keys(
            "Client",
            config.get("spec", {}),
            {"clientId"},
            {
                "parentId",
                "dacTags",
                "dacDefaultRead",
                "dacDefaultWrite",
                "dacDefaultExecute",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
                "macTags",
                "macWhitelistRead",
                "macWhitelistWrite",
                "macWhitelistExecute",
                "macBlacklistRead",
                "macBlacklistWrite",
                "macBlacklistExecute",
            },
        ):
            raise ValueError()
        self.config = config
        self.name = config["metadata"]["name"]
        self.remote = config["metadata"]["namespace"]

        parent_id = config["spec"].get("parentId")
        if parent_id:
            self.parent_ref_or_id = ResourceIdRef("Client", parent_id)

    def resource_ids(self):
        yield ResourceIdRef("Client", self.config["spec"]["clientId"])

    def id(self, resolver: "ResourceResolver"):
        return ResourceId("Client", resolver.link(self.config["spec"]["clientId"]))


class GroupResource(Resource):
    def __init__(self, config):
        self.kind = "Group"

        if not validate_keys(
            "Group",
            config.get("spec", {}),
            {"ownerId", "groupId", "remote"},
            {
                "public",
                "includeClients",
                "includeOrgs",
                "includeGroups",
                "excludeClients",
                "excludeOrgs",
                "excludeGroups",
            },
        ):
            raise ValueError
        self.config = config
        self.name = config["metadata"]["name"]
        self.parent_ref_or_id = ResourceIdRef("Client", config["spec"]["ownerId"])
        self.remote = config["spec"]["remote"]

    def resource_ids(self):
        yield ResourceIdRef("Group", self.config["spec"]["groupId"])


class LoopResource(Resource):
    def __init__(self, config):
        self.kind = "Loop"

        if not validate_keys(
            "loop",
            config.get("spec", {}),
            {"ownerId", "loopId", "remote"},
            {
                "dacTags",
                "dacDefaultRead",
                "dacDefaultWrite",
                "dacDefaultExecute",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
            },
        ):
            raise ValueError()
        self.config = config
        self.name = config["metadata"]["name"]
        self.parent_ref_or_id = ResourceIdRef("Client", config["spec"]["ownerId"])
        self.remote = config["spec"]["remote"]

    def resource_ids(self) -> ResourceIdRef:
        yield ResourceIdRef("Loop", self.config["spec"]["loopId"])

    def connection_info(self, resolver: "ResourceResolver"):
        loop_id = resolver.link(self.config["spec"]["loopId"])

        base = resolver.get_remote_config(self.remote)["loops"]
        rest_base, ws_base = _split_base_uri(base)
        path = "/loop/" + loop_id

        return LoopConnectionInfo(
            rest_info=ConnectionInfo(rest_base, path),
            ws_info=ConnectionInfo(ws_base, path),
        )


def _split_base_uri(base):
    if base.startswith("unix:"):
        rest_base = f"http+unix:{base}"
        ws_base = f"ws+unix:{base}"
    elif base.startswith("http:"):
        rest_base = base
        ws_base = base.replace("http:", "ws:", 1)
    elif base.startswith("https:"):
        rest_base = base
        ws_base = base.replace("https:", "wss:", 1)
    else:
        raise ValueError(f"Invalid base URI {base}")
    return rest_base, ws_base


class StreamResource(Resource):
    def __init__(self, config):
        self.kind = "Stream"

        if not validate_keys(
            "stream",
            config.get("spec", {}),
            {"loopId", "streamId", "remote"},
            {
                "dacTags",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
            },
        ):
            raise ValueError()
        self.config = config
        self.name = config["metadata"]["name"]
        self.parent_ref_or_id = ResourceIdRef("Loop", config["spec"]["loopId"])
        self.remote = config["spec"]["remote"]

    def resource_ids(self):
        return
        yield

    def loop(self, resolver: "ResourceResolver") -> LoopResource:
        return self.parent(resolver)

    def connection_info(self, resolver: "ResourceResolver") -> StreamConnectionInfo:
        stream_id = resolver.link(self.config["spec"]["streamId"])
        loop_info = self.loop(resolver).connection_info(resolver)
        group = None
        return loop_info.stream_connection_info(stream_id, group)


class ClusterResource(Resource):
    def __init__(self, config):
        self.kind = "Cluster"

        if not validate_keys(
            "cluster",
            config.get("spec", {}),
            {"ownerId", "clusterId", "remote"},
            {
                "dacTags",
                "dacDefaultRead",
                "dacDefaultWrite",
                "dacDefaultExecute",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
            },
        ):
            raise ValueError()
        self.config = config
        self.name = config["metadata"]["name"]
        self.parent_ref_or_id = ResourceIdRef("Client", config["spec"]["ownerId"])
        self.remote = config["spec"]["remote"]

    def resource_ids(self):
        yield ResourceIdRef("Cluster", self.config["spec"]["clusterId"])

    def connection_info(self, resolver: "ResourceResolver") -> ClusterConnectionInfo:
        cluster_id = resolver.link(self.config["spec"]["clusterId"])

        base_uris = resolver.get_remote_config(self.remote)

        configure_base = _split_base_uri(base_uris["clusters"])[0]
        configure_path = "/cluster/" + cluster_id

        connect_base = _split_base_uri(base_uris["loops"])[1]
        connect_path = "/cluster/" + cluster_id

        return ClusterConnectionInfo(
            configure_info=ConnectionInfo(configure_base, configure_path),
            connect_info=ConnectionInfo(connect_base, connect_path),
        )


class Spec:
    def compile(self):
        raise NotImplementedError()


class ResourceSpec(Spec):
    def __init__(self, config):
        self.config = config

    def compile(self):
        """Apply the given resource config."""
        items = self.config["spec"]
        error = False

        if not validate_keys(
            "spec",
            items,
            set(),
            {"groups", "clients", "loops", "streams", "clusters"},
        ):
            error = True

        for client in items.get("clients", []):
            try:
                yield ResourceSpec.compile_client(client)
            except ValueError as e:
                error = True
                continue

        for group in items.get("groups", []):
            try:
                yield ResourceSpec.compile_group(group)
            except ValueError as e:
                error = True
                continue

        for loop in items.get("loops", []):
            try:
                yield ResourceSpec.compile_loop(loop)
            except ValueError as e:
                error = True
                continue

        for stream in items.get("streams", []):
            try:
                yield ResourceSpec.compile_stream(stream)
            except ValueError as e:
                error = True
                continue

        for cluster in items.get("clusters", []):
            try:
                yield ResourceSpec.compile_cluster(cluster)
            except ValueError as e:
                error = True
                continue

        if error:
            raise ValueError()

    @classmethod
    def compile_client(cls, config: dict):
        if not validate_keys(
            "client",
            config,
            {"name"},
            {
                "namespace",
                "parent",
                "dacTags",
                "dacDefaultRead",
                "dacDefaultWrite",
                "dacDefaultExecute",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
                "macTags",
                "macWhitelistRead",
                "macWhitelistWrite",
                "macWhitelistExecute",
                "macBlacklistRead",
                "macBlacklistWrite",
                "macBlacklistExecute",
            },
        ):
            raise ValueError()

        client_config = {
            "apiVersion": "thatone.ai/v1",
            "kind": "Client",
            "metadata": {
                "name": config["name"],
                "namespace": Remote(
                    resourceKey=ResourceReference("Client", config["name"]),
                    remote=config.get("namespace"),
                ),
            },
            "spec": {
                "clientId": ClientId(config["name"]),
                "parentId": ClientId(config.get("parent"))
                if "parent" in config
                else None,
                "dacTags": config.get("dacTags", []),
                "dacDefaultRead": [
                    GroupId(x) for x in config.get("dacDefaultRead", [])
                ],
                "dacDefaultWrite": [
                    GroupId(x) for x in config.get("dacDefaultWrite", [])
                ],
                "dacDefaultExecute": [
                    GroupId(x) for x in config.get("dacDefaultExecute", [])
                ],
                "dacWhitelistRead": [
                    GroupId(x) for x in config.get("dacWhitelistRead", [])
                ],
                "dacWhitelistWrite": [
                    GroupId(x) for x in config.get("dacWhitelistWrite", [])
                ],
                "dacWhitelistExecute": [
                    GroupId(x) for x in config.get("dacWhitelistExecute", [])
                ],
                "dacBlacklistRead": [
                    GroupId(x) for x in config.get("dacBlacklistRead", [])
                ],
                "dacBlacklistWrite": [
                    GroupId(x) for x in config.get("dacBlacklistWrite", [])
                ],
                "dacBlacklistExecute": [
                    GroupId(x) for x in config.get("dacBlacklistExecute", [])
                ],
                "macTags": config.get("macTags", []),
                "macWhitelistRead": [
                    GroupId(x) for x in config.get("macWhitelistRead", [])
                ],
                "macWhitelistWrite": [
                    GroupId(x) for x in config.get("macWhitelistWrite", [])
                ],
                "macWhitelistExecute": [
                    GroupId(x) for x in config.get("macWhitelistExecute", [])
                ],
                "macBlacklistRead": [
                    GroupId(x) for x in config.get("macBlacklistRead", [])
                ],
                "macBlacklistWrite": [
                    GroupId(x) for x in config.get("macBlacklistWrite", [])
                ],
                "macBlacklistExecute": [
                    GroupId(x) for x in config.get("macBlacklistExecute", [])
                ],
            },
        }

        return ClientResource(client_config)

    @classmethod
    def compile_group(cls, config: dict):
        if not validate_keys(
            "group",
            config,
            {"name", "owner"},
            {
                "public",
                "includeClients",
                "includeOrgs",
                "includeGroups",
                "excludeClients",
                "excludeOrgs",
                "excludeGroups",
            },
        ):
            raise ValueError

        group_config = {
            "apiVersion": "thatone.ai/v1",
            "kind": "Group",
            "metadata": {
                "name": config["name"],
            },
            "spec": {
                "remote": Remote(
                    resourceKey=ResourceReference("Group", config["name"])
                ),
                "ownerId": ClientId(config["owner"]),
                "groupId": GroupId(config["name"]),
                "public": config.get("public", False),
                "includeClients": [
                    ClientId(x) for x in config.get("includeClients", [])
                ],
                "includeOrgs": [ClientId(x) for x in config.get("includeOrgs", [])],
                "includeGroups": [GroupId(x) for x in config.get("includeGroups", [])],
                "excludeClients": [
                    ClientId(x) for x in config.get("excludeClients", [])
                ],
                "excludeOrgs": [ClientId(x) for x in config.get("excludeOrgs", [])],
                "excludeGroups": [GroupId(x) for x in config.get("excludeGroups", [])],
            },
        }

        return GroupResource(group_config)

    @classmethod
    def compile_loop(cls, config: dict):
        if not validate_keys(
            "loop",
            config,
            {"name", "owner"},
            {
                "dacTags",
                "dacDefaultRead",
                "dacDefaultWrite",
                "dacDefaultExecute",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
            },
        ):
            raise ValueError()

        loop_config = {
            "apiVersion": "thatone.ai/v1",
            "kind": "Loop",
            "metadata": {
                "name": config["name"],
            },
            "spec": {
                "remote": Remote(resourceKey=ResourceReference("Loop", config["name"])),
                "ownerId": ClientId(config["owner"]),
                "loopId": LoopId(config["name"]),
                "dacTags": config.get("dacTags", []),
                "dacDefaultRead": [
                    GroupId(x) for x in config.get("dacDefaultRead", [])
                ],
                "dacDefaultWrite": [
                    GroupId(x) for x in config.get("dacDefaultWrite", [])
                ],
                "dacDefaultExecute": [
                    GroupId(x) for x in config.get("dacDefaultExecute", [])
                ],
                "dacWhitelistRead": [
                    GroupId(x) for x in config.get("dacWhitelistRead", [])
                ],
                "dacWhitelistWrite": [
                    GroupId(x) for x in config.get("dacWhitelistWrite", [])
                ],
                "dacWhitelistExecute": [
                    GroupId(x) for x in config.get("dacWhitelistExecute", [])
                ],
                "dacBlacklistRead": [
                    GroupId(x) for x in config.get("dacBlacklistRead", [])
                ],
                "dacBlacklistWrite": [
                    GroupId(x) for x in config.get("dacBlacklistWrite", [])
                ],
                "dacBlacklistExecute": [
                    GroupId(x) for x in config.get("dacBlacklistExecute", [])
                ],
            },
        }

        return LoopResource(loop_config)

    @classmethod
    def compile_stream(cls, config: dict):
        if not validate_keys(
            "stream",
            config,
            {"name", "loop"},
            {
                "dacTags",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
            },
        ):
            raise ValueError()

        stream_config = {
            "apiVersion": "thatone.ai/v1",
            "kind": "Stream",
            "metadata": {
                "name": config["name"],
            },
            "spec": {
                "remote": Remote(
                    resourceKey=ResourceReference("Stream", config["name"])
                ),
                "loopId": LoopId(config["loop"]),
                "streamId": config["name"],
                "dacTags": config.get("dacTags", []),
                "dacWhitelistRead": [
                    GroupId(x) for x in config.get("dacWhitelistRead", [])
                ],
                "dacWhitelistWrite": [
                    GroupId(x) for x in config.get("dacWhitelistWrite", [])
                ],
                "dacWhitelistExecute": [
                    GroupId(x) for x in config.get("dacWhitelistExecute", [])
                ],
                "dacBlacklistRead": [
                    GroupId(x) for x in config.get("dacBlacklistRead", [])
                ],
                "dacBlacklistWrite": [
                    GroupId(x) for x in config.get("dacBlacklistWrite", [])
                ],
                "dacBlacklistExecute": [
                    GroupId(x) for x in config.get("dacBlacklistExecute", [])
                ],
            },
        }

        return StreamResource(stream_config)

    @classmethod
    def compile_cluster(cls, config: dict):
        if not validate_keys(
            "cluster",
            config,
            {"name", "owner"},
            {
                "dacTags",
                "dacDefaultRead",
                "dacDefaultWrite",
                "dacDefaultExecute",
                "dacWhitelistRead",
                "dacWhitelistWrite",
                "dacWhitelistExecute",
                "dacBlacklistRead",
                "dacBlacklistWrite",
                "dacBlacklistExecute",
            },
        ):
            raise ValueError()

        cluster_config = {
            "apiVersion": "thatone.ai/v1",
            "kind": "Cluster",
            "metadata": {
                "name": config["name"],
            },
            "spec": {
                "remote": Remote(
                    resourceKey=ResourceReference("Cluster", config["name"])
                ),
                "ownerId": ClientId(config["owner"]),
                "clusterId": ClusterId(config["name"]),
                "dacTags": config.get("dacTags", []),
                "dacDefaultRead": [
                    GroupId(x) for x in config.get("dacDefaultRead", [])
                ],
                "dacDefaultWrite": [
                    GroupId(x) for x in config.get("dacDefaultWrite", [])
                ],
                "dacDefaultExecute": [
                    GroupId(x) for x in config.get("dacDefaultExecute", [])
                ],
                "dacWhitelistRead": [
                    GroupId(x) for x in config.get("dacWhitelistRead", [])
                ],
                "dacWhitelistWrite": [
                    GroupId(x) for x in config.get("dacWhitelistWrite", [])
                ],
                "dacWhitelistExecute": [
                    GroupId(x) for x in config.get("dacWhitelistExecute", [])
                ],
                "dacBlacklistRead": [
                    GroupId(x) for x in config.get("dacBlacklistRead", [])
                ],
                "dacBlacklistWrite": [
                    GroupId(x) for x in config.get("dacBlacklistWrite", [])
                ],
                "dacBlacklistExecute": [
                    GroupId(x) for x in config.get("dacBlacklistExecute", [])
                ],
            },
        }

        return ClusterResource(cluster_config)


class ResourcePile:
    def __init__(self, *paths: str, read_fully=False):
        self.read_fully = read_fully
        self.known_keys: dict[ResourceReference, dict[str, int]] = defaultdict(
            lambda: defaultdict(int)
        )
        self.config_sources: dict[ResourceReference, LocatedConfig] = {}
        self.compiled_resources: dict[ResourceReference, Resource] = {}

        self.add(*paths)

    def add(self, *paths: str):
        new_configs = list(
            itertools.chain(
                *[
                    self.read_resource_configs(file, read_fully=self.read_fully)
                    for file in paths
                ]
            )
        )
        new_keys = self.check_duplicates(new_configs)
        new_config_sources = self.map_config_locations(new_configs)
        new_compiled_resources = self.compile_configs(new_configs)

        self.known_keys = new_keys
        self.config_sources = new_config_sources
        self.compiled_resources = new_compiled_resources

    def read_resource_configs_file(
        self, file, read_fully=False
    ) -> Generator[LocatedConfig, None, None]:
        try:
            with open(file, "r") as f:
                configs = list(yaml.safe_load_all(f.read()))
        except Exception as e:
            print(f"Failed to read file {file}:", e.args)
            if read_fully:
                raise e
            return

        for i, config in enumerate(configs):
            if read_fully:
                yield LocatedConfig(file, i, config)
                continue

            if "apiVersion" not in config:
                print(f"Config in file {file} is missing apiVersion")
                continue
            if not isinstance(config["apiVersion"], str):
                print(
                    f"Config in file {file} has invalid apiVersion. It should be a string."
                )
                continue
            if not re.match(r"^[a-z0-9.-]+/[a-z0-9.-]+$", config["apiVersion"]):
                print(
                    f"Config in file {file} has invalid apiVersion. It should be in the format group/version. Group and version should be lowercase alphanumeric characters, dots or dashes."
                )
                continue
            if "kind" not in config:
                print(f"Config in file {file} is missing kind")
                continue
            if not isinstance(config["kind"], str):
                print(f"Config in file {file} has invalid kind. It should be a string.")
                continue
            if "metadata" not in config:
                print(f"Config in file {file} is missing metadata")
                continue
            if not isinstance(config["metadata"], dict):
                print(
                    f"Config in file {file} has invalid metadata. It should be a dict."
                )
                continue
            if "name" not in config["metadata"]:
                print(f"Config in file {file} is missing metadata.name")
                continue
            if not isinstance(config["metadata"]["name"], str):
                print(
                    f"Config in file {file} has invalid metadata.name. It should be a string."
                )
                continue

            if config["apiVersion"] != "thatone.ai/v1":
                print(
                    f"Skipping config in {file} with apiVersion {config['apiVersion']}."
                )
                continue
            if config["kind"] not in CONFIG_KINDS:
                print(f"Skipping config in {file} with kind {config['kind']}.")
                continue

            yield LocatedConfig(file, i, config)

    def read_resource_configs(self, dirpath, read_fully=False) -> List[LocatedConfig]:
        """Read the given file and return its contents."""
        dirpath = os.path.realpath(dirpath)
        if os.path.isfile(dirpath):
            return list(self.read_resource_configs_file(dirpath, read_fully=read_fully))

        result: List[LocatedConfig] = []
        for root, dirs, files in os.walk(dirpath):
            for filepath in files:
                if filepath.endswith(".yaml"):
                    fullpath = os.path.join(root, filepath)
                    result.extend(
                        self.read_resource_configs_file(fullpath, read_fully=read_fully)
                    )

        return result

    def check_duplicates(self, new_configs):
        new_keys = self.known_keys.copy()
        error = False

        for located_config in new_configs:
            if not is_itl_resource(located_config.config):
                continue

            for key in expand_keys(located_config.config):
                new_keys[key][located_config.path] += 1

        for key, file_counts in new_keys.items():
            if sum(file_counts.values()) > 1:
                error = True
                files_str = "\n  ".join(file_counts.keys())
                print(f"Error: Duplicate config {key} in files:\n  {files_str}")

        if error:
            print(
                "You can either remove each duplicate or apply the configs one at a time."
            )

        if error:
            raise ValueError()

        return new_keys

    def map_config_locations(self, prior_configs: List[LocatedConfig]):
        error = False
        new_config_sources = self.config_sources.copy()

        for located_config in prior_configs:
            if located_config.config.get("apiVersion") != "thatone.ai/v1":
                continue

            # try:
            #     resource = next(iter(self._compile_config(located_config)))
            # except ValueError as e:
            #     error = True
            #     continue

            # key = resource.reference
            key = ResourceReference(
                located_config.config["kind"], located_config.config["metadata"]["name"]
            )
            if key in new_config_sources:
                error = True
            else:
                new_config_sources[key] = located_config

        if error:
            raise ValueError()

        return new_config_sources

    def _compile_config(self, located_config: LocatedConfig) -> List["Resource"]:
        """Apply the given config."""
        kind = located_config.config["kind"]

        if kind == "ResourceSpec":
            spec = ResourceSpec(located_config.config)
            return list(spec.compile())
        elif kind == "ApiKey":
            return [ApiKeyResource(located_config.config)]
        elif kind == "Group":
            return [GroupResource(located_config.config)]
        elif kind == "Client":
            return [ClientResource(located_config.config)]
        elif kind == "Loop":
            return [LoopResource(located_config.config)]
        elif kind == "Stream":
            return [StreamResource(located_config.config)]
        elif kind == "Cluster":
            return [ClusterResource(located_config.config)]
        else:
            print("Unknown config kind", kind)
            raise ValueError(f"Unknown config kind {kind} in {located_config.path}")

    def compile_configs(self, new_configs: List[LocatedConfig]):
        error = False
        new_compiled_resources = self.compiled_resources.copy()

        for located_config in new_configs:
            try:
                generated_resources = self._compile_config(located_config)
                for resource in generated_resources:
                    key = resource.reference
                    if key in new_compiled_resources:
                        print("skipping", key, "since it was already generated")
                        continue
                    new_compiled_resources[key] = resource
            except ValueError as e:
                error = True
            except Exception as e:
                raise e

        if error:
            raise ValueError()

        return new_compiled_resources

    def link_resources(self, resolver: "ResourceResolver") -> dict[str, List[dict]]:
        unlinked_resources = self.compiled_resources
        prior_configs: Dict[
            ResourceReference, LocatedConfig
        ] = resolver.prior_resources.config_sources
        prior_resources: Dict[
            ResourceReference, Resource
        ] = resolver.prior_resources.compiled_resources

        error = False
        updated_configs: Dict[str, List[dict]] = defaultdict(list)
        updated_yaml_files = set()

        for located_config in prior_configs.values():
            updated_configs[located_config.path].append(located_config.config)

        for target_resource in unlinked_resources.values():
            reference = target_resource.reference
            if reference in prior_configs:
                location = prior_configs[reference]
                updated_configs[location.path][location.index] = target_resource.link(
                    resolver
                )
                updated_yaml_files.add(location.path)
            else:
                if target_resource.kind == "ApiKey":
                    target_file = resolver.create_secret_path(target_resource.config)
                else:
                    target_file = resolver.create_resource_path(target_resource.config)

                try:
                    updated_configs[target_file].append(target_resource.link(resolver))
                except ValueError as e:
                    print("Failed to generate config for", reference)
                    error = True
                    continue
                updated_yaml_files.add(target_file)

        if error:
            raise ValueError()

        return {x: updated_configs[x] for x in updated_yaml_files}

    def apply(
        self, prior_resources: "ResourcePile", resources_path: str, secrets_path: str
    ):
        resolver = ResourceResolver(prior_resources, self, resources_path, secrets_path)
        updated_configs = self.link_resources(resolver)

        for filepath, configs in updated_configs.items():
            dirname = os.path.dirname(filepath)
            os.makedirs(dirname, exist_ok=True)

            with open(filepath, "w") as outp:
                yaml.dump_all(
                    configs,
                    stream=outp,
                    default_flow_style=False,
                    sort_keys=False,
                    explicit_end=False,
                )

        for resource in self.compiled_resources.values():
            try:
                resource.apply(resolver)
            except Exception as e:
                print(f"Failed to apply {resource.reference}")
                print(e)
                raise

        for apikey in resolver.autogenerated_resources.values():
            path = resolver.create_secret_path(apikey.config)
            dirname = os.path.dirname(path)
            os.makedirs(dirname, exist_ok=True)
            with open(path, "w") as outp:
                yaml.dump(
                    apikey.link(resolver),
                    outp,
                    default_flow_style=False,
                    sort_keys=False,
                    explicit_end=False,
                )


class ResourceResolver:
    def __init__(
        self,
        prior_resources: ResourcePile,
        new_resources=None,
        resources_path=None,
        secrets_path=None,
    ):
        self.prior_resources: ResourcePile = prior_resources
        self.compiled_resources: dict[ResourceReference, Resource] = (
            new_resources.compiled_resources if new_resources else {}
        )
        self.autogenerated_resources: dict[ResourceReference, Resource] = {}
        self.secrets_path: Optional[str] = secrets_path
        self.resources_path: Optional[str] = resources_path
        self.generated_values: dict[ResourceReference, str] = {}
        self.id_to_reference: Dict[ResourceId, ResourceReference] = {}

        for reference, resource in self.prior_resources.compiled_resources.items():
            for resource_id in resource.resource_ids():
                self.id_to_reference[resource_id.resolve(self)] = reference

        for reference, resource in self.compiled_resources.items():
            for resource_id in resource.resource_ids():
                self.id_to_reference[resource_id.resolve(self)] = reference

    def create_path(self, folder, config):
        """Create the path for the given resource."""
        if folder == None:
            raise ValueError("No folder provided to the ResourceResolver")
        prefix = os.path.join(folder, config["kind"], config["metadata"]["name"])
        suffix = 0
        while True:
            result = prefix + (suffix or "") + ".yaml"
            if not os.path.exists(result):
                return result
            suffix += 1

    def create_secret_path(self, config):
        return self.create_path(self.secrets_path, config)

    def create_resource_path(self, config):
        return self.create_path(self.resources_path, config)

    def random_id(self):
        random_bytes = os.urandom(20)
        return "".join("{:02x}".format(byte) for byte in random_bytes)

    def get_generated_id(self, configKey: ConfigKey) -> str:
        created = (
            self.compiled_resources.get(configKey.reference)
            if self.compiled_resources
            else None
        )
        if created != None:
            spec: dict[str, Any] = created.config["spec"]
            result = spec.get(configKey.key)
            if result and not isinstance(result, DeferredValue):
                return result

        generated = self.generated_values.get(configKey.reference)
        if generated:
            return generated

        existing = self.prior_resources.compiled_resources.get(configKey.reference)
        if existing:
            result = existing.config["spec"].get(configKey.key)
            if result:
                return result

        new_id = f"{configKey.reference.name}-{self.random_id()}"
        self.generated_values[configKey.reference] = new_id
        return new_id

    def store_generated_apikey(self, cause: ClientResource, apikey_config):
        client_name = cause.config["metadata"]["name"]
        suffix = 0
        while True:
            if suffix == 0:
                candidate_name = f"{client_name}-apikey"
            else:
                candidate_name = f"{client_name}-apikey-{suffix}"

            apikey_reference = ResourceReference("ApiKey", candidate_name)
            if apikey_reference in self.compiled_resources:
                suffix += 1
                continue

            if apikey_reference in self.prior_resources.compiled_resources:
                suffix += 1
                continue

            break

        apikey = ApiKeyResource(
            {
                "apiVersion": apikey_config["apiVersion"],
                "kind": apikey_config["kind"],
                "metadata": {
                    "name": apikey_reference.name,
                    "namespace": cause.config["metadata"]["namespace"],
                },
                "spec": apikey_config["spec"],
            }
        )

        self.autogenerated_resources[apikey.reference] = apikey
        return apikey

    def link(self, id: Union[DeferredValue, ResourceId, ResourceIdRef, str]) -> str:
        if isinstance(id, str):
            return id
        elif isinstance(id, DeferredValue):
            return id.link(self)
        elif isinstance(id, ResourceId):
            return id.id
        elif isinstance(id, ResourceIdRef):
            return id.resolve(self).id
        else:
            raise ValueError(f"Invalid id {id}")

    def apikeys_for_resource(self, resource: "Resource"):
        while True:
            if isinstance(resource, ClientResource):
                resource_id = ResourceId(
                    "Client", self.link(resource.config["spec"]["clientId"])
                )
                return self.apikeys_for_client(resource_id)
            if resource.parent(self) == None:
                break
            resource = resource.parent(self)

    def get_resources_by_type(self, kind) -> Generator["Resource", None, None]:
        for candidate, resource in self.compiled_resources.items():
            if candidate.kind == kind:
                yield resource

        for candidate, value in self.prior_resources.compiled_resources.items():
            if candidate.kind == kind:
                yield value

        for candidate, value in self.autogenerated_resources.items():
            if candidate.kind == kind:
                yield value

    def get_remote_config(self, remote: Union[Remote, str]) -> "Resource":
        remote = self.link(remote)
        for candidate in self.get_resources_by_type("Remote"):
            if candidate.config["metadata"]["name"] == remote:
                return candidate.config["spec"]

        return DEFAULT_REMOTES[remote]

    def apikeys_for_client(
        self, client_id: ResourceId
    ) -> Generator["ApiKeyResource", None, None]:
        for apikey in self.get_resources_by_type("ApiKey"):
            assert isinstance(apikey, ApiKeyResource)

            if not apikey.valid:
                continue

            config = apikey.config
            if self.link(config["spec"]["clientId"]) != client_id.id:
                continue

            yield apikey

        parent_id = self.get_client_parent_id(client_id)
        if parent_id == None:
            return

        for apikey in self.apikeys_for_client(parent_id):
            yield apikey

    def get_resource_by_reference(self, reference: ResourceReference):
        created = self.compiled_resources.get(reference)
        if created:
            return created

        return self.prior_resources.compiled_resources.get(reference)

    def require_resource_by_reference(self, reference: ResourceReference):
        resource = self.get_resource_by_reference(reference)
        if resource == None:
            print("Error: Could not find resource", reference)
            raise ValueError()
        return resource

    def get_client_parent_id(self, client_id: ResourceId):
        client_reference = self.id_to_reference.get(client_id)
        if client_reference == None:
            return None

        assert client_reference is not None

        resource = self.get_resource_by_reference(client_reference)
        if resource == None:
            return None

        config = resource.config["spec"]
        if config.get("parentId") != None:
            return ResourceId("Client", self.link(config["parentId"]))

        return None



import argparse
import asyncio
from contextlib import contextmanager
import os
from urllib.parse import urlparse
import requests

import fire
import yaml

from .itl import Itl
from .clusters import ClusterOperations


@contextmanager
def edit_config(config):
    if os.path.exists(config):
        with open(config, "r") as f:
            config_data = yaml.safe_load(f)
    else:
        config_data = {
            "apiVersion": "itllib/v1",
            "kind": "ItlConfig",
            "metadata": {
                "name": "itl-config",
            },
            "spec": {},
        }

    yield config_data

    with open(f"{config}.temp", "w") as f:
        yaml.dump(config_data, f)
    os.rename(f"{config}.temp", config)


def register_db_with_notifier(db, endpoint, notification_url):
    print("Setting up database notifications for", db)
    url = f"https://{endpoint}/{db}/notification"
    params = {"endpoint": notification_url + "/event"}
    headers = {"accept": "application/json"}
    response = requests.post(url, headers=headers, params=params, data="")
    print("Done")


class ConfigUpdater:
    def add_stream(self, name, loop, config="./config.yaml"):
        with edit_config(config) as config_data:
            stream_data = {
                "name": name,
                "loop": loop,
            }

            streams = config_data["spec"].setdefault("streams", [])
            for stream in streams:
                if stream["name"] != name:
                    continue

                overwrite = input(f"The stream {name} already exists. Overwrite? [y/N]")
                if overwrite.lower() != "y":
                    print(
                        f"Skipping {name}. If you want to write the config elsewhere, provide a different --config path."
                    )
                    return

                stream.clear()
                stream.update(stream_data)
                return

            streams.append(stream_data)

    def add_cluster(
        self, name, database, stream=None, config="./config.yaml", secrets="./secrets"
    ):
        itl = Itl()
        itl._apply_config(config, secrets)
        if database not in itl._old_clusters:
            raise ValueError(f"Could not find database {database} in {secrets}")

        dbOps = itl._old_clusters[database]
        if stream:
            if stream not in itl._streams:
                raise ValueError(f"Could not find stream {stream} in {secrets}")
            streamObj = itl._streams.get(stream)
        else:
            streamObj = None

        with edit_config(config) as config_data:
            cluster_data = {
                "name": name,
                "database": database,
            }

            if stream:
                cluster_data["eventStream"] = stream

            clusters = config_data["spec"].setdefault("clusters", [])
            for cluster in clusters:
                if cluster["name"] != name:
                    continue

                overwrite = input(
                    f"The cluster {name} already exists. Overwrite? [y/N]"
                )
                if overwrite.lower() != "y":
                    print(
                        f"Skipping {name}. If you want to write the config elsewhere, provide a different --config path."
                    )
                    return

                asyncio.run(ClusterOperations(dbOps, stream, streamObj, "").create())
                cluster.clear()
                cluster.update(cluster_data)
                return

            asyncio.run(ClusterOperations(dbOps, stream, streamObj, "").create())
            clusters.append(cluster_data)

    def dump_endpoints(self, config="./config.yaml", secrets="./secrets"):
        itl = Itl()
        itl._apply_config(config, secrets)

        result = {}

        for name, stream in itl._streams.items():
            result.setdefault("streams", {})[name] = {
                "post": stream.send_url,
                "websocket": stream.connect_url,
            }

        for name, db in itl._old_clusters.items():
            result.setdefault("databases", {})[name] = {
                "url": f"{db.endpoint_url}/{db.name}",
            }

        print(yaml.dump(result))


class Configurator:
    config = ConfigUpdater()

    def create_loop(self, id, endpoint, name, secrets="./secrets"):
        os.makedirs(secrets, exist_ok=True)
        secret_path = os.path.join(secrets, f"{id}.yaml")
        if os.path.exists(secret_path):
            overwrite = input(
                f"The loop config will be stored in {secret_path}. Overwrite {secret_path}? [y/N]"
            )
            if overwrite.lower() != "y":
                print(
                    f"Skipping {id}. If you want to write the config elsewhere, provide a different --secrets path."
                )
                return

        with open(secret_path, "w") as f:
            f.write(
                f"""
apiVersion: itllib/v1
kind: LoopSecret
metadata:
  name: {id}
  creationTimestamp: null
spec:
  loopName: {name}
  authenticationType: PASSWORD
  secretBasicAuth:
    endpoint: {endpoint}
    username: usernameGoesHere
    password: passwordGoesHere
  protocols: [basicAuth]
"""
            )

    def create_database(self, id, endpoint, name, notifier=None, secrets="./secrets"):
        os.makedirs(secrets, exist_ok=True)
        secret_path = os.path.join(secrets, f"{id}.yaml")
        if os.path.exists(secret_path):
            overwrite = input(
                f"The database config will be stored in {secret_path}. Overwrite {secret_path}? [y/N]"
            )
            if overwrite.lower() != "y":
                print(
                    f"Skipping {id}. If you want to write the config elsewhere, provide a different --secrets path."
                )
                return

        with open(secret_path, "w") as f:
            f.write(
                f"""
apiVersion: itllib/v1
kind: DatabaseSecret
metadata:
  name: {id}
  creationTimestamp: null
spec:
  databaseName: {name}
  notifier: {notifier}
  authenticationType: PASSWORD
  secretBasicAuth:
    endpoint: {endpoint}
    username: usernameGoesHere
    password: passwordGoesHere
  protocols: [basicAuth]
"""
            )

        if notifier:
            register_db_with_notifier(name, endpoint, notifier)

    def apply(self, config):
        raise NotImplementedError()


if __name__ == "__main__":
    fire.Fire(Configurator())



import asyncio
from collections import defaultdict
import inspect
import threading
import typing
from glob import glob
import traceback
from urllib.parse import urlparse

import websockets
import json
import aiohttp
import yaml
import requests

from itllib.resources import (
    ResourcePile,
    ClientResource,
    ApiKeyResource,
    GroupResource,
    LoopResource,
    ResourceReference,
    ResourceResolver,
    StreamResource,
    ClusterResource,
)

from .piles import BucketOperations, PileOperations
from .clusters import ClusterOperations
from .loops import (
    ConnectionInfo,
    LoopOperations,
    StreamConnectionInfo,
    StreamOperations,
)


class Namespace:
    pass


def _get_expected_arguments(func):
    signature = inspect.signature(func)
    return set(signature.parameters.keys())


def _get_argument_type_hints(func):
    type_hints = typing.get_type_hints(func)
    signature = inspect.signature(func)
    argument_type_hints = {}

    for name, param in signature.parameters.items():
        if name in type_hints:
            argument_type_hints[name] = type_hints[name]

    return argument_type_hints


class Itl:
    def __init__(self, *configs, client=None) -> None:
        # User-specified handlers
        self._data_handlers = {}
        self._controllers = {}

        # Async stuff
        self._connection_thread = None
        self._connection_looper = None
        self._connection_tasks = asyncio.Queue()
        self._callback_thread = None
        self._callback_looper = None
        self._callback_tasks = asyncio.Queue()

        # Old stuff, to be removed
        self._stop = False

        # Resources
        self._secrets = {}
        self._streams: dict[str, StreamOperations] = {}
        self._buckets = {}
        self._piles = {}
        self._clusters: dict[str, ClusterOperations] = {}
        self._loops: dict[str, LoopOperations] = {}

        # Stream interactions
        self._downstreams = {}
        self._upstreams = {}
        self._start_persistent_tasks = {}
        self._downstream_queues = defaultdict(asyncio.Queue)
        self._started_streams = set()
        self._start_ephemeral_tasks = []
        self._post_connection_tasks = []

        self._resource_pile = ResourcePile()
        self._resolver = ResourceResolver(self._resource_pile)
        self._keys = defaultdict(lambda: None)

        self._apply_config(*configs, client=client)

    def _apply_config(self, *files, client=None):
        self._resource_pile.add(*files)
        self._resolver = ResourceResolver(self._resource_pile)

        # Get apikeys for the client
        if client != None:
            client_ref = ResourceReference("Client", client)
            client_resource = self._resolver.get_resource_by_reference(client_ref)
            assert isinstance(client_resource, ClientResource)
            client_id = client_resource.id(self._resolver)

            for apikey in self._resolver.apikeys_for_client(client_id):
                remote = apikey.get_remote(self._resolver)
                self._keys[remote] = apikey.key(self._resolver)

        for resource in self._resource_pile.compiled_resources.values():
            if isinstance(resource, LoopResource):
                remote = resource.get_remote(self._resolver)
                connection_info = resource.connection_info(self._resolver)
                self._loops[resource.name] = LoopOperations(
                    connection_info, self._keys[remote]
                )
            elif isinstance(resource, StreamResource):
                remote = resource.get_remote(self._resolver)
                connection_info = resource.connection_info(self._resolver)
                self._streams[resource.name] = StreamOperations(
                    connection_info, self._keys[remote]
                )
            elif isinstance(resource, ClusterResource):
                remote = resource.get_remote(self._resolver)
                connection_info = resource.connection_info(self._resolver)
                self._clusters[resource.name] = ClusterOperations(
                    connection_info, self._keys[remote]
                )
                self._streams["cluster/" + resource.name] = StreamOperations(
                    StreamConnectionInfo(None, connection_info.connect_info),
                    self._keys[remote],
                )

    def attach_pile_prefix(self, pile, name):
        return self._piles[pile].prefix + name

    def object_download(self, pile, key=None, notification=None, attach_prefix=False):
        if key == None and notification == None:
            raise ValueError("Exactly one of key or event must be provided")
        if key != None and notification != None:
            raise ValueError("Only one of key or event can be provided")

        if notification != None:
            key = notification["key"]

        pile_ops = self._piles[pile]

        if attach_prefix:
            key = f"{pile_ops.prefix or ''}{key}"

        return pile_ops.get(key)

    def object_upload(
        self, pile, key, file_descriptor, metadata={}, attach_prefix=False
    ):
        pile_ops = self._piles[pile]

        if attach_prefix:
            key = f"{pile_ops.prefix or ''}{key}"

        return pile_ops.put(key, file_descriptor, metadata)

    def object_delete(self, pile, key=None, attach_prefix=False):
        pile_ops = self._piles[pile]

        if attach_prefix:
            key = f"{pile_ops.prefix or ''}{key}"

        return pile_ops.delete(key)

    async def cluster_create(self, cluster, data):
        return await self._clusters[cluster].create_resource(data)

    async def cluster_read_all(
        self, cluster, group=None, version=None, kind=None, name=None, utctime=None
    ):
        return await self._clusters[cluster].read_all_resources(
            group, version, kind, name, utctime
        )

    async def cluster_read(self, cluster, group, version, kind, name):
        return await self._clusters[cluster].read_resource(group, version, kind, name)

    async def cluster_patch(self, cluster, data):
        return await self._clusters[cluster].patch_resource(data)

    async def cluster_update(self, cluster, data):
        return await self._clusters[cluster].update_resource(data)

    async def cluster_apply(self, cluster, data):
        return await self._clusters[cluster].apply_resource(data)

    async def cluster_delete(self, cluster, group, version, kind, name):
        return await self._clusters[cluster].delete_resource(group, version, kind, name)

    def cluster_controller(self, cluster, group, version, kind, name, validate=True):
        cluster_obj = self._clusters[cluster]
        return cluster_obj.control_resource(
            group, version, kind, name, validate=validate
        )
        # yield controller

    def _get_url(self, identifier) -> str:
        if identifier in self._streams:
            return self._streams[identifier].connect_url
        else:
            return identifier

    def _ensure_stream_connection(self, streams):
        """
        Update the upstream tasks based on the provided streams. If an upstream task for a
        given identifier already exists, it is skipped. If the looper isn't initialized,
        a task to attach the downstream is created. If the looper is initialized, a new downstream task
        is scheduled to run asynchronously.

        Args:
        - streams (List[str]): List of stream identifiers to be processed.

        Returns:
        None
        """
        for identifier in streams:
            # Skip creating a task if it already exists
            if identifier in self._start_persistent_tasks:
                continue

            # Check if the Itl is already running
            if not self._connection_looper:
                task = self._attach_stream, (identifier,)
                self._start_persistent_tasks[identifier] = task
            else:
                self._connection_looper.call_soon_threadsafe(
                    self._connection_tasks.put_nowait,
                    lambda: self._schedule_stream_task_unsafe(identifier),
                )

    def _schedule_stream_task_unsafe(self, identifier):
        """
        Schedules an upstream task for the given identifier if it doesn't already exist.

        Args:
        - identifier (str): Identifier for the stream.

        Returns:
        None
        """
        if identifier in self._start_persistent_tasks:
            return

        task = self._attach_stream, (identifier,)
        self._start_persistent_tasks[identifier] = task
        asyncio.create_task(self._attach_stream(identifier))

    def ondata(self, stream=None, loop=None, streamId=None):
        if stream == None:
            loop_info = self._loops[loop]
            stream = f'stream/{loop}/{streamId}'
            if stream not in self._streams:
                self._streams[stream] = StreamOperations(
                    loop_info.connect_info.stream_connection_info(streamId),
                    loop_info.apikey,
                )
        if stream not in self._upstreams:
            self._ensure_stream_connection([stream])

        def decorator(func):
            self._data_handlers.setdefault(stream, []).append(func)
            return func

        return decorator

    def controller(
        self,
        cluster,
        group=None,
        version=None,
        kind=None,
        name=None,
        validate=True,
    ):
        cluster_obj = self._clusters[cluster]
        stream = "cluster/" + cluster

        def decorator(func):
            async def controller_wrapper(*args, **event):
                operations = self.cluster_controller(
                    cluster,
                    event["group"],
                    event["version"],
                    event["kind"],
                    event["name"],
                    validate=validate,
                )
                async with operations:
                    try:
                        await func(operations)
                    except Exception as e:
                        print(
                            f"Error in controller {func.__name__}: {traceback.format_exc()}"
                        )

            @self.ondata(stream)
            async def event_handler(*args, **event):
                if event["event"] != "queue":
                    return
                if group and event["group"] != group:
                    return
                if version and event["version"] != version:
                    return
                if kind and event["kind"] != kind:
                    return

                asyncio.create_task(controller_wrapper(*args, **event))

            self._controllers.setdefault(cluster, []).append(func)

            async def check_queue():
                for queued_op in await cluster_obj.read_queue(
                    group, version, kind, name
                ):
                    asyncio.create_task(controller_wrapper(**queued_op))

            self.onconnect(check_queue)
            return func

        return decorator

    def onconnect(self, func):
        if self._callback_looper:
            self._callback_looper.call_soon_threadsafe(
                self._callback_tasks.put_nowait, (func, ())
            )
        else:
            self._start_ephemeral_tasks.append((func, ()))

        return func

    async def _process_upstream_messages(self):
        def exec_callback(handler, args, kwargs):
            try:
                if inspect.iscoroutinefunction(handler):
                    asyncio.create_task(handler(*args, **kwargs))
                else:
                    handler(*args, **kwargs)
            except Exception as e:
                print(f"Error in handler {handler.__name__}: {traceback.format_exc()}")

        def process_message(identifier, serialized_data):
            message = json.loads(serialized_data)
            # TODO: Run all handlers in parallel
            for handler in self._data_handlers[identifier]:
                if isinstance(message, dict):
                    args = []
                    kwargs = message
                else:
                    args = [message]
                    kwargs = {}

                exec_callback(handler, args, kwargs)

        while True:
            task = await self._callback_tasks.get()

            if self._stop:
                break

            if task == None:
                continue

            identifier, serialized_data = task

            try:
                if isinstance(identifier, str):
                    process_message(identifier, serialized_data)
                else:
                    exec_callback(identifier, serialized_data, {})
            except asyncio.CancelledError:
                pass
            except Exception as e:
                print(f"Error in message processing: {traceback.format_exc()}")

    async def _post_stream_message(self, url, message):
        # call HTTP POST on key, passing message as data
        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=json.dumps(message)) as response:
                response.raise_for_status()

    def stream_send(self, key=None, message=None, loop=None, streamId=None):
        if key == None:
            loop_info = self._loops[loop]
            key = f'stream/{loop}/{streamId}'
            if key not in self._streams:
                self._streams[key] = StreamOperations(
                    loop_info.connect_info.stream_connection_info(streamId),
                    loop_info.apikey,
                )

        if key not in self._streams and (
            key.startswith("http://") or key.startswith("https://")
        ):
            asyncio.run_coroutine_threadsafe(
                self._post_stream_message(key, message), self._connection_looper
            )
            return

        self._ensure_stream_connection([key])

        if self._connection_looper:
            self._connection_looper.call_soon_threadsafe(
                self._downstream_queues[key].put_nowait, message
            )
        else:
            task = self._downstream_queues[key].put, (message,)
            self._connection_looper.call_soon_threadsafe(
                self._downstream_queues[key].put_nowait, task
            )

    def stream_send_sync(self, key, message):
        if key.startswith("http://") or key.startswith("https://"):
            # call HTTP POST on key, passing message as data
            requests.post(url=key, json=json.dumps(message))
            return

        self._ensure_stream_connection([key])

        if self._connection_looper:
            self._downstream_queues[key].put_nowait(message)
        else:
            task = self._downstream_queues[key].put, (message,)
            self._downstream_queues[key].put_nowait(task)

    def _requeue(self, identifier, message):
        if message == None:
            return

        old_queue = self._downstream_queues[identifier]
        new_queue = asyncio.Queue()
        if message != None:
            new_queue.put_nowait(message)
        while not old_queue.empty():
            new_queue.put_nowait(old_queue.get_nowait())

        self._downstream_queues[identifier] = new_queue

    async def _attach_stream(self, identifier):
        if identifier in self._started_streams:
            return

        self._started_streams.add(identifier)
        state = Namespace()
        state.message = None

        if identifier not in self._streams:
            raise NotImplementedError("Streams must be defined in the config")

        apikey = self._streams[identifier].apikey

        async def send_message():
            state.message = (
                state.message or await self._downstream_queues[identifier].get()
            )

            if self._stop:
                self._requeue(identifier, state.message)
                return False

            serialized_data = json.dumps(state.message)

            try:
                await self._streams[identifier].send(serialized_data)
            except websockets.exceptions.ConnectionClosedError:
                return False
            except websockets.exceptions.ConnectionClosedOK:
                return False

            state.message = None
            return True

        async def recv_message():
            try:
                serialized_data = await self._streams[identifier].recv()
            except websockets.exceptions.ConnectionClosedError:
                return False
            except websockets.exceptions.ConnectionClosedOK:
                return False

            # If there are no data handlers for this identifier, skip processing
            if identifier not in self._data_handlers:
                return True

            self._callback_looper.call_soon_threadsafe(
                self._callback_tasks.put_nowait, (identifier, serialized_data)
            )

            return True

        backoff_time = 0
        tasks = None

        while not self._stop:
            try:
                if not tasks:
                    tasks = [
                        asyncio.create_task(asyncio.sleep(0)),
                        asyncio.create_task(asyncio.sleep(0)),
                    ]

                ws_url = self._get_url(identifier)
                if apikey:
                     ws_url = ws_url + "/apikey/" + apikey

                async with websockets.connect(ws_url) as websocket:
                    backoff_time = 0
                    self._streams[identifier].socket = websocket

                    while True:
                        done, pending = await asyncio.wait(
                            tasks, return_when=asyncio.FIRST_COMPLETED
                        )
                        if self._stop:
                            return

                        connection_closed = False

                        for completed in done:
                            if completed.result() == False:
                                connection_closed = True
                                tasks = None
                            elif completed == tasks[0]:
                                tasks[0] = asyncio.create_task(send_message())
                            elif completed == tasks[1]:
                                tasks[1] = asyncio.create_task(recv_message())

                        if connection_closed:
                            break

            except websockets.exceptions.ConnectionClosedOK:
                pass
            except websockets.exceptions.ConnectionClosedError:
                pass

            if self._stop:
                return

            # Backoff before reconnecting
            backoff_time = await self._exponential_backoff(backoff_time)

    async def _exponential_backoff(self, current_backoff_time):
        """Sleeps the process for an exponential backoff time."""
        await asyncio.sleep(2**current_backoff_time)
        # Return the next backoff time, capped at 2**7 seconds
        return min(current_backoff_time + 1, 7)

    async def _connect(self):
        connection_tasks = []
        for fn, args in self._start_persistent_tasks.values():
            asyncio.create_task(fn(*args))
        for fn, args in self._start_ephemeral_tasks:
            connection_tasks.append(fn(*args))

        await asyncio.gather(*connection_tasks)

        post_connection_tasks = []
        for fn, args in self._post_connection_tasks:
            post_connection_tasks.append(fn(*args))

        await asyncio.gather(*post_connection_tasks)

        # if self._post_connect:
        #     await self._post_connect()

    def start(self):
        self._connection_thread = threading.Thread(
            target=self._handle_connections_in_thread
        )
        self._connection_thread.start()

        self._callback_thread = threading.Thread(
            target=self._handle_callbacks_in_thread
        )
        self._callback_thread.start()

    def _handle_connections_in_thread(self):
        self._connection_looper = looper = asyncio.new_event_loop()
        asyncio.set_event_loop(looper)
        looper.set_debug(True)
        looper.run_until_complete(self._start_routine())
        looper.close()

    def _handle_callbacks_in_thread(self):
        self._callback_looper = looper = asyncio.new_event_loop()
        asyncio.set_event_loop(looper)
        looper.set_debug(True)
        looper.run_until_complete(self._process_upstream_messages())
        looper.close()

    def stop(self):
        self._stop = True
        self._callback_looper.call_soon_threadsafe(
            self._callback_tasks.put_nowait, None
        )
        self._connection_looper.call_soon_threadsafe(
            self._connection_tasks.put_nowait, None
        )

    async def _start_routine(self):
        self._looper = asyncio.get_event_loop()
        tasks = []

        await self._connect()
        await asyncio.gather(*tasks)

        while True:
            task = await self._connection_tasks.get()
            if self._stop:
                break
            task()

        for queue in self._downstream_queues.values():
            queue.put_nowait(None)

        close_tasks = []
        for stream in self._downstreams.values():
            close_tasks.append(stream.close())

        for stream in self._upstreams.values():
            close_tasks.append(stream.close())

        asyncio.gather(*close_tasks)
